import logging
import typing as tp

from .base import BaseSyncWorker, SyncError
from ..exceptions import ResponseError

logger = logging.getLogger(__name__)


class HTTPSyncWorker(BaseSyncWorker):

    def sync_pathpoints(self, data: tp.List[dict]):
        try:
            self.api.post('/v1/device/pathpoints', json=data, timeout=40)
        except ResponseError as e:
            raise SyncError(e.is_no_link(), e.status_code // 100 == 4) from e
        logger.debug('Successfully synchronized %s pathpoints', len(data))

    def sync_logs(self, data: tp.List[dict]):
        try:
            self.api.put('/v1/device/device_logs', json=data, timeout=40)
        except ResponseError as e:
            if e.status_code // 100 == 4:
                logger.error('Tried to sync logs %s but resulted in %s', data, e.status_code)
            raise SyncError(e.is_no_link(), e.status_code // 100 == 4) from e

    def __init__(self, device: 'SMOKDevice'):
        super().__init__(device, False)
        self.api = device.api

    def close(self):
        pass



import os
import pickle
import typing as tp

from satella.coding import silence_excs, wraps

from smok.extras.macros_database.in_memory import InMemoryMacroDatabase
from smok.macro import Macro


def always_sync(fun):
    @wraps(fun)
    def inner(self, *args, **kwargs):
        try:
            return fun(self, *args, **kwargs)
        finally:
            getattr(self, '_PicklingMacroDatabase__sync')()

    return inner


class PicklingMacroDatabase(InMemoryMacroDatabase):
    """
    An example persistent macro database

    :param path: path to pickle file
    """

    def __init__(self, path):
        super().__init__()
        self.__path = path

        if os.path.exists(self.__path):
            with open(self.__path, 'rb') as f_in, silence_excs(pickle.PickleError):
                self.macros_to_execute, self.executions_to_sync = pickle.load(f_in)

    def __sync(self):
        with open(self.__path, 'wb') as f_out:
            pickle.dump((self.macros_to_execute, self.executions_to_sync), f_out,
                        pickle.HIGHEST_PROTOCOL)

    @always_sync
    def set_macros(self, macros: tp.List[Macro]) -> None:
        super().set_macros(macros)

    @always_sync
    def get_macros(self) -> tp.List[Macro]:
        return super().get_macros()

    @always_sync
    def notify_macro_executed(self, macro_id: str, timestamp: int) -> None:
        return super().notify_macro_executed(macro_id, timestamp)

    def get_done_macros(self) -> tp.Iterator[tp.Tuple[str, int]]:
        return super().get_done_macros()

    @always_sync
    def notify_macro_synced(self, macro_id: str, timestamp: int) -> None:
        return super().notify_macro_synced(macro_id, timestamp)



import logging
from concurrent.futures import Future

import minijson
from satella.coding import wraps, for_argument, silence_excs, rethrow_as
from satella.coding.optionals import Optional
from satella.exceptions import Empty
from satella.time import ExponentialBackoff

from ..orders import Order

import typing as tp
import select
from satella.coding.concurrent import TerminableThread

from ..exceptions import DataStreamSyncFailed, ConnectionFailed
from ..protocol import NGTTHeaderType
from .connection import NGTTSocket

logger = logging.getLogger(__name__)


def must_be_connected_else_raise(fun):
    @wraps(fun)
    def outer(self, *args, **kwargs):
        if not self.connected:
            raise ConnectionFailed(True)
        return fun(self, *args, **kwargs)

    return outer


class NGTTConnection(TerminableThread):
    """
    An interface to NGTT, also a thread maintaining connection in the background.

    Note that instantiating this object is the same as calling start. You do not need to call
    start on this object after you initialize it.

    :param cert_file: path to file with certificate. This file should contain only the device
        certificate, attaching entire certificate chain is not required.
    :param key_file: path to private key
    :param on_new_order: a callable taking only a single argument and returning nothing, the
        callable to call when a new order appears. Note that you have to call
        either :meth:`~ngtt.orders.Order.acknowledge` or :meth:`~ngtt.orders.Order.nack` for each
        received order. Leave at default (None) if orders are not meant to be fetched.

    :ivar connected (bool) is connection opened
    """

    def __init__(self, cert_file: str, key_file: str,
                 on_new_order: tp.Optional[tp.Callable[[Order], None]] = None):
        super().__init__(name='ngtt uplink')
        self.on_new_order = on_new_order
        self.cert_file = cert_file
        self.stopped = False
        self.key_file = key_file
        self.current_connection = None
        self.start()

    def stop(self, wait_for_completion: bool = True):
        """
        Stop this thread and the connection

        :param wait_for_completion: whether to wait for thread to terminate
        """
        if self.stopped:
            return
        self.terminate()
        if wait_for_completion:
            self.join()
        self.stopped = True

    def close(self):
        """
        Alias for :meth:`~ngtt.uplink.NGTTConnection.stop`.
        """
        self.stop()
        if self.current_connection is not None:
            self.current_connection.close()
            self.current_connection = None
            self.op_id_to_op = {}

    @property
    @silence_excs(AttributeError, returns=False)
    def connected(self) -> bool:
        """Are we connected to target server?"""
        return self.current_connection.connected

    def connect(self):
        if self.connected:
            return
        eb = ExponentialBackoff(1, 30, self.safe_sleep)
        while not self.terminating and not self.connected:
            try:
                self.current_connection = NGTTSocket(self.cert_file, self.key_file)
                self.current_connection.connect()
                if self.on_new_order:
                    self.current_connection.send_frame(0, NGTTHeaderType.FETCH_ORDERS)
            except ConnectionFailed as e:
                logger.warning('Failure reconnecting', exc_info=e)
                eb.failed()
                eb.sleep()

            if self.terminating:
                return

    @must_be_connected_else_raise
    @for_argument(None, minijson.dumps)
    def sync_pathpoints(self, data) -> Future:
        """
        Try to synchronize pathpoints.

        This will survive multiple reconnection attempts.

        :param data: exactly the same thing that you would submit to POST
        at POST https://api.smok.co/v1/device/
        :return: a Future telling you whether this succeeds or fails. If succeeds, it will
            end with a result of None, if fails, it will end with a
            :class:`~ngtt.exceptions.DataStreamSyncFailed`
        """
        fut = Future()
        fut.set_running_or_notify_cancel()
        try:
            tid = self.current_connection.id_assigner.allocate_int()
        except Empty as e:
            logger.error('Ran out of IDs with a NGTT connection', exc_info=e)
            raise ConnectionFailed(False, 'Ran out of IDs to assign')

        self.current_connection.send_frame(tid, NGTTHeaderType.DATA_STREAM, data)
        self.current_connection.futures[tid] = fut
        return fut

    def inner_loop(self):
        self.current_connection.try_ping()
        ccon = [self.current_connection]
        rx, wx, ex = select.select(ccon,
                                   ccon if self.current_connection.wants_write else [], [],
                                   1)
        if wx:
            with rethrow_as(ConnectionResetError, ConnectionFailed):
                self.current_connection.try_send()
        if not rx:
            return
        frame = self.current_connection.recv_frame()
        if frame is None:
            return
        if frame.packet_type == NGTTHeaderType.PING:
            self.current_connection.got_ping()
        elif frame.packet_type == NGTTHeaderType.ORDER:
            try:
                data = frame.real_data
            except ValueError:
                logger.error('Received invalid JSON over the wire')
                raise ConnectionFailed(False, 'Got invalid JSON')
            order = Order(data, frame.tid, self.current_connection)
            self.on_new_order(order)
        elif frame.packet_type in (
                NGTTHeaderType.DATA_STREAM_REJECT, NGTTHeaderType.DATA_STREAM_CONFIRM):
            tid = frame.tid
            if tid in self.current_connection.futures:
                self.current_connection.id_assigner.mark_as_free(tid)
                # Assume it's a data stream running
                fut = self.current_connection.futures.pop(tid)

                if frame.packet_type == NGTTHeaderType.DATA_STREAM_CONFIRM:
                    fut.set_result(None)
                elif frame.packet_type == NGTTHeaderType.DATA_STREAM_REJECT:
                    fut.set_exception(DataStreamSyncFailed())
            else:
                logger.info('This was an unknown confirmation')

    def loop(self) -> None:
        try:
            self.connect()
        except ConnectionFailed as e:
            logger.warning('Failure during connect', exc_info=e)
            return

        if self.terminating:
            return

        try:
            self.inner_loop()
        except ConnectionFailed as e:
            logger.warning('Connection failed', exc_info=e)
            self.cleanup()

    def cleanup(self):
        Optional(self.current_connection).close()
        self.current_connection = None

    @must_be_connected_else_raise
    @for_argument(None, minijson.dumps)
    def stream_logs(self, data: tp.List) -> None:
        """
        Stream logs to the server

        This will work on a best-effort basis.

        :param data: the same thing that you would PUT /v1/device/device_logs
        """
        self.current_connection.send_frame(0, NGTTHeaderType.LOGS, data)



import typing as tp
from abc import abstractmethod, ABCMeta


class SyncError(Exception):
    """
    An error has happened during the sync
    """

    def __init__(self, is_no_link: bool = True, is_clients_fault: bool = False):
        self._is_no_link = is_no_link
        self._is_clients_fault = is_clients_fault

    def is_no_link(self):
        return self._is_no_link

    def is_clients_fault(self):
        return self._is_clients_fault


class BaseSyncWorker(metaclass=ABCMeta):
    """
    Base class for synchronizing archives and logs

    :ivar has_async_orders: (bool) whether this can asynchronously issue orders. If it can,
        you need to take care to inject received orders appropriately.
    """

    def __init__(self, device: 'SMOKDevice', has_async_orders: bool):
        self.device = device
        self.has_async_orders = has_async_orders

    @abstractmethod
    def sync_logs(self, data: tp.List[dict]):
        """
        :raises SyncError: on failure
        """

    @abstractmethod
    def close(self):
        ...

    @abstractmethod
    def sync_pathpoints(self, data: tp.List[dict]):
        """
        Synchronize the pathpoint data

        :raises SyncError: on failure
        """



import minijson

import struct
import typing as tp

from satella.coding.structures import HashableIntEnum, ReprableMixin

from .exceptions import InvalidFrame


class NGTTHeaderType(HashableIntEnum):
    PING = 0  #: a PING, sent by the client, server responds alike
    ORDER = 1  #: sent by the server
    ORDER_CONFIRM = 2  #: sent by the client
    LOGS = 3  #: sent by the client
    DATA_STREAM = 4  #: sent by the client
    DATA_STREAM_CONFIRM = 5  #: sent by the server
    DATA_STREAM_REJECT = 6  #: sent by the server
    ORDER_REJECT = 9  #: sent by the client, currently unused (by the client)
    FETCH_ORDERS = 10  #: a request by the client to fetch orders


STRUCT_LHH = struct.Struct('>LHH')


class NGTTFrame(ReprableMixin):
    """
    A basic NGTT protocol frame. It's big endian. It starts with a header of:

    uint32 data_size
    uint16 order_id
    uint16 packet_type
    bytes[data_size] data

    Data is represented using minijson.

    :param tid: order ID
    :param packet_type: type of the packet
    :param data: data. Any class that has a __bytes__ property will do.
    """
    _REPR_FIELDS = 'tid', 'packet_type', 'data'

    def __init__(self, tid: int, packet_type: NGTTHeaderType, data=b''):
        self.tid = tid
        self.packet_type = packet_type
        self.data = bytes(data)

    def __str__(self) -> str:
        return repr(self)

    @property
    def real_data(self) -> tp.Union[dict, list]:
        """
        Unserialize the data

        :return: JSON unserialized data
        :raises ValueError: invalid MiniJSON
        """
        return minijson.loads(self.data)

    def __len__(self):
        return STRUCT_LHH.size + len(self.data)

    def __bytes__(self):
        return STRUCT_LHH.pack(len(self.data), self.tid, self.packet_type.value) + self.data

    @classmethod
    def from_bytes(cls, b: tp.Union[bytes, bytearray]) -> 'NGTTFrame':
        length, tid, h_type = STRUCT_LHH.unpack(b[:STRUCT_LHH.size])
        if h_type > NGTTHeaderType.DATA_STREAM_REJECT.value:
            raise InvalidFrame('Unrecognized packet type %s' % (h_type,))
        return NGTTFrame(tid, NGTTHeaderType(h_type), b[STRUCT_LHH.size:STRUCT_LHH.size + length])

    @classmethod
    def from_buffer(cls, buffer: bytearray) -> tp.Optional[tp.Tuple[int, 'NGTTFrame']]:
        """
        Try to construct a frame from a buffer.

        If successful, a tuple will be output of amount of bytes consumed and an instance of
        :class:`~ngtt.protocol.NGTTFrame`.
        If failure, None will be returned
        """
        if len(buffer) < STRUCT_LHH.size:
            return None
        length, tid, h_type = STRUCT_LHH.unpack(buffer[:STRUCT_LHH.size])
        if not length:
            return STRUCT_LHH.size, NGTTFrame(tid, NGTTHeaderType(h_type))
        if len(buffer) < length + STRUCT_LHH.size:
            return None
        return length + STRUCT_LHH.size, NGTTFrame(tid, NGTTHeaderType(h_type),
                                                   buffer[STRUCT_LHH.size:STRUCT_LHH.size + length])


def env_to_hostname(env: int) -> str:
    return {0: 'api.smok.co',
            1: 'api.test.smok-serwis.pl'}.get(env, 'rapid-rs')



import copy
import typing as tp

from satella.coding import Monitor
from satella.coding.typing import Number

from .base import BaseDataToSynchronize, BasePathpointDatabase, ValueOrExcept
from ...exceptions import OperationFailedError, NotReadedError, OperationFailedReason
from ...pathpoint import PathpointValueType


class InMemoryDataToSynchronize(BaseDataToSynchronize):
    def negative_acknowledge(self) -> None:
        pass

    __slots__ = 'in_memory', 'data'

    def __init__(self, in_memory: 'InMemoryPathpointDatabase', data: tp.List):
        self.in_memory = in_memory
        self.data = data

    def to_json(self) -> tp.List:
        return self.data

    def acknowledge(self) -> None:
        timestamps_for_pathpoints = {}  # type: tp.Dict[str, Number]
        for pp in self.data:
            max_ts = max(v['timestamp'] for v in pp['values'])
            timestamps_for_pathpoints[pp['path']] = max_ts

        for pathpoint, timestamp in timestamps_for_pathpoints.items():
            self.in_memory.confirm_synced_up_to(pathpoint, timestamp)


class InMemoryPathpointDatabase(BasePathpointDatabase, Monitor):
    def get_archive_data(self, pathpoint: str, start: tp.Optional[Number],
                         stop: tp.Optional[Number]) -> tp.Iterator[tp.Tuple[Number, ValueOrExcept]]:
        return []

    @Monitor.synchronized
    def get_current_value(self, pathpoint: str) -> tp.Tuple[Number, PathpointValueType]:
        """
        Get the current value for given pathpoint

        :param pathpoint: name of the pathpoint
        :return: a tuple of (timestamp, value or exception)
        :raises NotReadedError: pathpoint has no last value
        """
        if pathpoint not in self.pathpoints:
            raise NotReadedError()

        lv = self.pathpoints[pathpoint][-1]
        if 'error_code' in lv:
            raise OperationFailedError(OperationFailedReason(lv['error_code']), lv['timestamp'])
        else:
            return lv['timestamp'], lv['value']

    @Monitor.synchronized
    def get_data_to_sync(self) -> tp.Optional[BaseDataToSynchronize]:
        if not self.pathpoints:
            return None
        data_to_sync = []
        for pathpoint, values in self.pathpoints.items():
            data_to_sync.append({'path': pathpoint, 'values': copy.copy(values)})
        return InMemoryDataToSynchronize(self, data_to_sync)

    @Monitor.synchronized
    def on_new_data(self, pathpoint: str, timestamp: Number, value_or_exception: ValueOrExcept):
        if isinstance(value_or_exception, OperationFailedError):
            datapoint = {'timestamp': timestamp, 'error_code': value_or_exception.reason.value}
        else:
            datapoint = {'timestamp': timestamp, 'value': value_or_exception}

        if pathpoint not in self.pathpoints:
            self.pathpoints[pathpoint] = [datapoint]
        else:
            if int(self.pathpoints[pathpoint][0]['timestamp']) < int(timestamp):
                self.pathpoints[pathpoint].append(datapoint)

    @Monitor.synchronized
    def confirm_synced_up_to(self, pathpoint: str, timestamp: Number) -> None:
        if pathpoint not in self.pathpoints:
            return
        for i, data in enumerate(self.pathpoints[pathpoint]):
            if data['timestamp'] > timestamp:
                break
        else:
            i = None

        if i is None:
            del self.pathpoints[pathpoint]
        else:
            self.pathpoints[pathpoint] = self.pathpoints[pathpoint][i:]

    def __init__(self):
        self.pathpoints = {}
        Monitor.__init__(self)



import typing as tp
from smok.extras import BasePredicateDatabase


class InMemoryPredicateDatabase(BasePredicateDatabase):
    def __init__(self):
        self.predicates = {}

    def get_all_predicates(self) -> tp.List[tp.Dict]:
        return list(self.predicates.values())

    def update_predicate(self, v: tp.Dict) -> None:
        self.predicates[v['predicate_id']] = v

    def set_new_predicates(self, v: tp.List[tp.Dict]):
        self.predicates = {}
        for dct in v:
            self.predicates[dct['predicate_id']] = dct



import logging
import typing as tp
from abc import ABCMeta, abstractmethod

logger = logging.getLogger(__name__)


class BAOBDigest:
    """
    A descriptor of a BAOB

    :ivar key: key of this BAOB (str)
    :ivar version: version of this BAOB (int)
    """

    __slots__ = ('key', 'version')

    def __init__(self, key: str, version: int):
        self.key = key
        self.version = version


class BaseBAOBDatabase(metaclass=ABCMeta):
    @abstractmethod
    def get_baob_value(self, key: str) -> bytes:
        """
        Return a value of a particular BAOB

        :raises KeyError: BAOB does not exist
        """

    @abstractmethod
    def delete_baob(self, key: str) -> None:
        """
        Delete particular BAOB

        :raises KeyError: BAOB does not exist
        """

    @abstractmethod
    def get_baob_version(self, key: str) -> int:
        """
        Return a version of particular BAOB

        :raises KeyError: BAOB does not exist
        """

    @abstractmethod
    def get_all_keys(self) -> tp.Iterator[str]:
        """
        Stream all keys of BAOBs stored locally
        """

    @abstractmethod
    def set_baob_value(self, key: str, data: bytes, version: int):
        """
        Store a particular version of a BAOB within the database
        """

    def check_consistency(self) -> None:
        """
        Check that for all keys that are returned by `get_all_keys` version can be
        obtained via `get_baob_version`.

        If not, delete such a key and log an ERROR.

        Called by constructor of :class:`~smok.client.SMOKDevice`
        """
        for key in self.get_all_keys():
            try:
                self.get_baob_version(key)
            except KeyError:
                logger.error(
                    'Key %s mentioned in all keys but cannot get a version for it, deleting', key)
                self.delete_baob(key)



class SensorWriteEvent:
    """
    An event describing a situation wherein a sensor was written

    :param timestamp: in milliseconds
    """
    __slots__ = ('timestamp', 'who', 'hr_sensor', 'hr_value',
                 'fqts', 'value', 'reason')

    def __hash__(self):
        return hash(self.timestamp)

    def __eq__(self, other):
        return self.timestamp == other.timestamp and self.fqts == other.fqts

    def __init__(self, timestamp: int, who: str, hr_sensor: str, hr_value: str,
                 fqts: str, value: str, reason: str):
        self.timestamp = timestamp
        self.who = who
        self.hr_sensor = hr_sensor
        self.hr_value = hr_value
        self.fqts = fqts
        self.value = value
        self.reason = reason

    def to_json(self):
        return {'timestamp': self.timestamp,
                'who': self.who, 'hr_sensor': self.hr_sensor, 'hr_value': self.hr_value,
                'fqts': self.fqts, 'value': self.value, 'reason': self.reason}



import queue
import logging
import typing as tp

from satella.coding import queue_get
from satella.coding.concurrent import TerminableThread
from satella.coding.decorators import retry
from satella.time import ExponentialBackoff

from ..sync_workers.base import SyncError

MAX_SYNC_AT_ONCE = 50
MAX_LOG_BUFFER_SIZE = 20000
logger = logging.getLogger(__name__)


class LogPublisherThread(TerminableThread):
    def __init__(self, device: 'SMOKDevice'):
        super().__init__(name='log publisher')
        self.device = device
        self.queue = queue.Queue()
        self.waiter = ExponentialBackoff()

    def get_all_messages(self, starting_msg: dict) -> tp.List[dict]:
        msgs = [starting_msg]
        while self.queue.qsize() and len(msgs) < MAX_SYNC_AT_ONCE:
            msgs.append(self.queue.get())
        return msgs

    @queue_get('queue', timeout=5)
    def loop(self, msg) -> None:
        self.safe_sleep(1)  # wait for other records to become available
        while not self.device.allow_sync:
            while self.queue.qsize() > MAX_LOG_BUFFER_SIZE:
                self.queue.get()
        if self._terminating:
            return
        msgs = self.get_all_messages(msg)
        self.sync(msgs)

    @retry(3, exc_classes=SyncError)
    def sync(self, lst: tp.List[dict]) -> None:
        try:
            try:
                self.device.sync_worker.sync_logs(lst)
                self.device.on_successful_sync()
                self.waiter.success()
            except Exception as e:
                logger.error('exception %s', e, exc_info=e)
                raise
        except SyncError as e:
            self.waiter.failed()
            self.waiter.sleep()
            if e.is_no_link():
                self.device.on_failed_sync()
            raise

    def cleanup(self) -> None:
        if self.device.allow_sync:
            while self.queue.qsize() > 0:
                msgs = self.get_all_messages(self.queue.get())
                self.sync(msgs)



from .archives_and_macros import ArchivingAndMacroThread
from .communicator import CommunicatorThread
from .executor import OrderExecutorThread
from .log_publisher import LogPublisherThread

__all__ = ['CommunicatorThread', 'OrderExecutorThread', 'ArchivingAndMacroThread',
           'LogPublisherThread']



import os
import pickle

from .in_memory import InMemorySensorWriteDatabase
from smok.sensor import SensorWriteEvent


class PicklingSensorWriteDatabase(InMemorySensorWriteDatabase):
    def __init__(self, pickle_addr: str):
        super().__init__()
        self.__path = pickle_addr
        if os.path.exists(pickle_addr):
            try:
                with open(pickle_addr, 'rb') as f_in:
                    self.events = pickle.load(f_in)
            except IOError:
                self.events = set()
        else:
            self.events = set()

    def sync(self):
        with open(self.__path, 'wb') as f_out:
            pickle.dump(self.events, f_out, pickle.HIGHEST_PROTOCOL)

    def add_sw(self, event: SensorWriteEvent):
        super().add_sw(event)
        self.sync()

    def on_synced_sw(self, event: SensorWriteEvent):
        super().on_synced_sw(event)
        self.sync()



import pickle
import typing as tp
import uuid

from satella.coding import update_key_if_true
from satella.coding.structures import HashableIntEnum
from satella.json import JSONAble
from satella.time import time_as_int


class Color(HashableIntEnum):
    """
    Event severity
    """
    WHITE = 0  #: least severe event
    YELLOW = 1  #: event of medium severity
    RED = 2  #: most severe event


class Event(JSONAble):
    """
    An object representing a single event in the SMOK system.

    :ivar uuid: event UUID (str) after being synced with the server
    :ivar provisional_uuid: UUID (str) assigned by the client after it's opening
    :ivar started_on: timestamp of event beginning, in seconds (int)
    :ivar ended_on: timestamp of event ending in seconds, or None if not ended (tp.Optional[int])
    :ivar color: event color (:class:`~smok.predicate.Color`)
    :ivar is_point: is the event a point one? (bool)
    :ivar token: token (str)
    :ivar group: event group (str)
    :ivar message: human-readable message (str)
    :ivar handled_by: user that handles this event (tp.Optional[str])
    :ivar metadata: event metadata (tp.Dict[str, str])
    """
    __slots__ = ('uuid', 'provisional_uuid', 'started_on', 'ended_on',
                 'color', 'is_point', 'token', 'group', 'message', 'handled_by',
                 'metadata')

    def to_pickle(self) -> bytes:
        """
        :return: pickled self
        """
        return pickle.dumps(self, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def from_pickle(cls, y: bytes) -> 'Event':
        """
        Load an event from a pickle

        :param y: pickled Event
        :type y: bytes
        :return: unpickled Event
        :rtype: Event
        """
        return pickle.loads(y)

    def get_uuid(self) -> str:
        if self.uuid:
            return self.uuid
        else:
            return self.provisional_uuid

    def uuid_matches(self, uuid_: str) -> bool:
        return uuid_ == self.uuid or uuid_ == self.provisional_uuid

    def __eq__(self, other: 'Event') -> bool:
        return self.get_uuid() == other.get_uuid()

    def __hash__(self) -> int:
        return hash(self.get_uuid())

    def __init__(self, uuid_: tp.Optional[str], started_on: tp.Optional[int],
                 ended_on: tp.Optional[int],
                 color: Color, is_point: bool, token: str, group: str, message: str,
                 handled_by: tp.Optional[str], metadata: tp.Dict[str, str]):
        self.uuid = uuid_
        if self.uuid is None:
            self.provisional_uuid = uuid.uuid4().hex
        else:
            self.provisional_uuid = None
        self.started_on = started_on or time_as_int()
        self.ended_on = ended_on
        self.color = color
        self.is_point = is_point
        self.token = token
        self.group = group
        self.message = message
        self.handled_by = handled_by
        self.metadata = metadata

    def to_json(self) -> dict:
        """
        Convert self to JSON representation
        """
        dct = {
            'started_on': self.started_on,
            'color': self.color.value,
            'alarm_type': 1 if self.is_point else 0,
            'token': self.token,
            'group': self.group,
            'message': self.message,
            'metadata': self.metadata
        }
        update_key_if_true(dct, 'uuid', self.uuid)
        update_key_if_true(dct, 'handled_by', self.handled_by)
        update_key_if_true(dct, 'ended_on', self.ended_on)
        return dct

    def is_closed(self) -> bool:
        """
        :return: is given event closed?
        """
        if self.is_point:
            return True
        return self.ended_on is not None

    @classmethod
    def from_json(cls, dct: dict) -> 'Event':
        """
        Restore self from a JSON representation
        """
        return Event(dct.get('uuid'), dct['started_on'], dct.get('ended_on'),
                     Color(dct['color']), dct['alarm_type'] == 1, dct['token'],
                     dct['group'], dct['message'], dct['metadata'])



from smok.exceptions import SMOKClientError


class NGTTError(SMOKClientError):
    """Base class for NGTT exceptions"""
    pass


class ConnectionFailed(NGTTError):
    """
    The request failed due to a problem with connection

    :ivar is_due_to_no_internet: was this error caused by the internet not being there?
    """

    def __init__(self, is_due_to_no_internet: bool = False, *args):
        super().__init__(*args)
        self.is_due_to_no_internet = is_due_to_no_internet


class DataStreamSyncFailed(NGTTError):
    """
    Server declined to sync data, as it's invalid.
    """


class InvalidFrame(NGTTError):
    """
    An invalid frame has been sent over the wire
    """



import typing as tp
from abc import ABCMeta, abstractmethod

from smok.sensor import Sensor


class BaseSensorDatabase(metaclass=ABCMeta):
    def __init__(self):
        self.device = None

    def on_register(self, device: 'SMOKDevice') -> None:
        """
        Called by SMOKDevice upon registering this database.

        Called by SMOKDevice's constructor.
        """
        self.device = device

    @abstractmethod
    def get_sensor(self, fqts: str) -> Sensor:
        """
        Return a sensor

        Called by user threads and possibly predicates (so also CommunicatorThread)

        :param fqts: fqts for target sensor, always in canonical form
        :raises KeyError: sensor not defined
        """

    @abstractmethod
    def on_sensors_sync(self, sensors: tp.List[Sensor]):
        """
        Sensors have just been synchronized, this is the entire list.

        Called by communicator threads.
        Will never be called if dont_do_pathpoints is enabled.
        """

    @abstractmethod
    def get_all_sensors(self) -> tp.Iterator[Sensor]:
        """
        Return all sensors stored in the database.

        Called by user threads and possibly predicates (so also CommunicatorThread).
        """



import logging
import time

from ngtt.orders import Order
from smok.pathpoint.orders import sections_from_list
from ngtt.uplink import NGTTConnection
from smok.sync_workers.http import HTTPSyncWorker

logger = logging.getLogger(__name__)


class NGTTSyncWorker(HTTPSyncWorker):
    """
    Note that this is ready to start when .executor on the device is ready
    """

    def __init__(self, device: 'SMOKDevice'):
        super().__init__(device)
        self.connection = NGTTConnection(device.temp_file_for_cert,
                                         device.temp_file_for_key,
                                         None if self.device.dont_do_orders else self.process_orders)
        self.has_async_orders = True
        time.sleep(5)  # Give it some time to connect

    def process_orders(self, orders: Order):
        sections = sections_from_list(orders.data)

        def confirm(fut):
            orders.acknowledge()

        sections[-1].future.add_done_callback(confirm)
        self.device.executor.queue.put_many(sections)

    def close(self):
        self.connection.close()



import logging
import json
import typing as tp

import minijson
import requests
from satella.files import read_in_file

from smok.basics import Environment
from smok.exceptions import ResponseError

logger = logging.getLogger(__name__)


class RequestsAPI:
    __slots__ = 'environment', 'base_url', 'cert'

    def __init__(self, device):
        self.environment = device.environment
        self.base_url = device.url

        # Prepare the certificate
        if self.environment == Environment.PRODUCTION:
            self.cert = device.temp_file_for_cert, device.temp_file_for_key
        elif self.environment == Environment.STAGING:
            self.cert = read_in_file(device.cert[0], 'utf-8').replace('\r\n', '\n').replace('\n',
                                                                                            '\t')
        else:
            self.cert = read_in_file('tests/dev.testing.crt', 'utf-8').replace('\r\n',
                                                                               '\n').replace('\n',
                                                                                             '\t')

    def request(self, request_type: str, url: str,
                direct_response: bool = False, **kwargs) -> tp.Union[dict, tp.Tuple[bytes, dict]]:
        """
        :param request_type: type of request, in lowercase
        :param url: URL to contact
        :param direct_response: if True then return a tuple of (
            response as is, headers), else it's JSON
        :param minijson: send provided dictionary as a MiniJSON

        :raises ResponseError: something went wrong
        """
        headers = kwargs.pop('headers', {})
        if 'json' in kwargs:
            kwargs['data'] = minijson.dumps(kwargs.pop('json'))
            headers['Content-Type'] = 'application/minijson'

        op = getattr(requests, request_type)
        try:
            if self.environment == Environment.PRODUCTION:
                resp = op(self.base_url + url, cert=self.cert, headers=headers, **kwargs)
            else:
                headers['X-SSL-Client-Certificate'] = self.cert
                resp = op(self.base_url + url, headers=headers, **kwargs)
        except requests.RequestException as e:
            logger.error('Requests error %s', e)
            raise ResponseError(599, 'Requests error: %s' % (str(e),)) from e

        if resp is None and self.environment == Environment.LOCAL_DEVELOPMENT:
            # special case for CI
            if direct_response:
                return '', 200
            else:
                return {}

        if resp.status_code not in (200, 201):
            if direct_response:
                return resp.content, resp.headers
            try:
                error_json = resp.json()['status']
            except (json.decoder.JSONDecodeError, ValueError):
                try:
                    error_json = resp.text
                except AttributeError:
                    error_json = repr(resp.content)
            raise ResponseError(resp.status_code, error_json)
        if direct_response:
            return resp.content, resp.headers
        else:
            try:
                return resp.json()
            except json.decoder.JSONDecodeError as e:
                raise ResponseError(resp.status_code, resp.content) from e

    def get(self, url: str, **kwargs):
        return self.request('get', url, **kwargs)

    def post(self, url: str, **kwargs):
        return self.request('post', url, **kwargs)

    def put(self, url: str, **kwargs):
        return self.request('put', url, **kwargs)

    def patch(self, url: str, **kwargs):
        return self.request('patch', url, **kwargs)

    def delete(self, url: str, **kwargs):
        return self.request('delete', url, **kwargs)



import copy
import logging
import queue
import typing as tp
from concurrent.futures import wait, Future

from satella.coding import queue_get
from satella.coding.concurrent import TerminableThread, call_in_separate_thread
from satella.coding.decorators import retry
from satella.instrumentation import Traceback
from satella.time import time_ms, measure

from smok.exceptions import ResponseError, NotReadedError, OperationFailedError
from smok.extras.pp_database.base import BasePathpointDatabase
from smok.pathpoint.orders import Section, WriteOrder, ReadOrder, MessageOrder, Disposition, \
    SysctlOrder, Order
from smok.pathpoint.pathpoint import Pathpoint

logger = logging.getLogger(__name__)


def on_read_completed_factory(pp: Pathpoint) -> tp.Callable[[Future], None]:
    def on_read_completed(fut: Future):
        ts = time_ms()
        if fut.exception() is None:
            res = fut.result()
            if res is None:
                return
        else:
            res = fut.exception()
            if not isinstance(res, OperationFailedError):
                try:
                    raise res
                except Exception as e:
                    f = Traceback().pretty_format()
                    logger.error('got %s while processing a read, stack trace is %s', e, f)
                    return
            if isinstance(res, NotReadedError):
                logger.error('A read future for %s returned NotReadedError, this is invalid, '
                             'ignoring',
                             pp.name)
                return
            res.timestamp = ts
        pp.set_new_value(ts, res)

    return on_read_completed


def on_write_completed(fut: Future):
    if fut.exception() is not None:
        exc = fut.exception()
        if not isinstance(exc, OperationFailedError):
            try:
                raise exc
            except Exception as e:
                f = Traceback().pretty_format()
                logger.error('got %s while processing a write, stack trace is %s, '
                             'assuming the write got through', e, f)
                return


class OrderExecutorThread(TerminableThread):
    def __init__(self, device, order_queue: queue.Queue, data_to_sync: BasePathpointDatabase,
                 wait_before_startup: int = 0):
        super().__init__(name='order executor')
        self.queue = order_queue
        self.device = device
        self.data_to_sync = data_to_sync
        self.wait_before_startup = wait_before_startup

    def prepare(self) -> None:
        if self.wait_before_startup:
            self.safe_sleep(self.wait_before_startup)

    def process_orders(self, orders) -> tp.List[Order]:
        """
        Do some orders and return those that need to be retried
        """
        futures_to_complete = []
        orders_to_complete = []
        for order in orders:
            if isinstance(order, (WriteOrder, ReadOrder)):
                try:
                    pathpoint = self.device.get_pathpoint(order.pathpoint)
                except KeyError:
                    continue

                if isinstance(order, WriteOrder):
                    if not order.is_valid():
                        continue
                    fut = pathpoint.on_write(order.value, order.advise)
                    if fut is None:
                        continue
                    fut.add_done_callback(on_write_completed)
                elif isinstance(order, ReadOrder):
                    if pathpoint.can_read():
                        try:
                            fut = pathpoint.on_read(order.advise)  # type: Future
                            if not isinstance(fut, Future):
                                if fut is not None:
                                    pathpoint.set_new_value(time_ms(), fut)
                            else:
                                fut.add_done_callback(on_read_completed_factory(pathpoint))
                        except OperationFailedError as e:
                            pathpoint.set_new_value(time_ms(), e)
                        except Exception as e:
                            logger.error('Read operation for %s raised %s, ignoring',
                                         pathpoint.name, e, exc_info=e)
                    continue
            elif isinstance(order, MessageOrder):

                @call_in_separate_thread()
                @retry(6, ResponseError)
                def execute_a_message(uuid: str) -> Future:
                    self.device.api.post('/v1/device/orders/message/' + uuid)

                fut = execute_a_message(order.uuid)
            elif isinstance(order, SysctlOrder):
                self.device.execute_sysctl(order.op_type, order.op_args)
            else:
                continue
            orders_to_complete.append(order)
            futures_to_complete.append(fut)

        new_futures_to_complete = copy.copy(futures_to_complete)
        while new_futures_to_complete and not self.terminating:
            new_futures_to_complete = wait(new_futures_to_complete, 5)[1]

        if self.terminating:
            return []

        orders_to_retry = []
        for future, order in zip(futures_to_complete, orders_to_complete):
            if isinstance(future.exception(), (ResponseError, OperationFailedError)):
                if order.fail():
                    orders_to_retry.append(order)
        return orders_to_retry

    def execute_a_section(self, section: Section) -> None:
        from smok.client import SMOKDevice

        # Do we need to sync all sections?
        if section.disposition == Disposition.CANNOT_JOIN:
            self.device.sync_sections(lambda: self.terminating)

        if self.terminating:
            return

        orders = section.orders

        if section.mark_as_being_executed():
            if self.device.__class__.execute_section is SMOKDevice.execute_section:
                with measure() as measurement:
                    while orders and not self.terminating:
                        orders = self.process_orders(orders)

                time_to_wait = section.max_wait()
                if time_to_wait is not None:
                    if measurement() < time_to_wait:
                        self.safe_sleep(time_to_wait - measurement())

            else:
                self.device.execute_section(section)
        section.mark_as_done()

    @queue_get('queue', 5)
    def loop(self, section: Section):
        if self.queue.qsize():
            next_section = self.queue.peek()  # type: Section
            if section.is_joinable() and next_section.is_joinable():
                section += self.queue.get()

        self.execute_a_section(section)



import typing as tp
from abc import ABCMeta, abstractmethod


class BasePredicateDatabase(metaclass=ABCMeta):
    """
    One predicate is a dict obtained from it's :meth:`~smok.predicate.BaseStatistic.to_kwargs`.

    All of these methods are called by CommunicatorThread, and
    :meth:`~smok.extras.BasePredicateDatabase.get_all_predicates` by SMOKDevice during
    initialization.
    """

    @abstractmethod
    def get_all_predicates(self) -> tp.List[tp.Dict]:
        """
        Return all predicates
        """

    @abstractmethod
    def update_predicate(self, v: tp.Dict) -> None:
        """
        A single predicate has been changed.

        Note when this is changed there can be no doubt that the predicate has changed. This is
        verified by CommunicatorThread.
        """

    @abstractmethod
    def set_new_predicates(self, v: tp.List[tp.Dict]):
        """
        Called upon a successful predicate synchronization from the server
        """



import typing as tp
from abc import ABCMeta, abstractmethod

from smok.sensor import SensorWriteEvent


class BaseSensorWriteSynchronization:
    """
    At most a single instance of this object will be kept alive.

    Must be boolable - return False if this is empty
    """
    __slots__ = ('events', 'swdb')

    def __init__(self, events: tp.List[SensorWriteEvent], swdb: 'BaseSensorWriteDatabase'):
        self.events = events
        self.swdb = swdb

    def __bool__(self):
        return bool(self.events)

    def to_json(self):
        return [event.to_json() for event in self.events]

    def ack(self):
        for event in self.events:
            self.swdb.on_synced_sw(event)

    def nack(self):
        for event in self.events:
            self.swdb.on_sync_sw_failed(event)


class BaseSensorWriteDatabase(metaclass=ABCMeta):
    def on_sync_sw_failed(self, event: SensorWriteEvent):
        """
        Called when given event fails to sync.

        This will never be called if dont_do_pathpoints is True
        """

    def on_synced_sw(self, event: SensorWriteEvent):
        """
        Called when given event becomes synced

        This will never be called if dont_do_pathpoints is True
        """

    @abstractmethod
    def add_sw(self, event: SensorWriteEvent):
        """
        Add a new event to synchronize
        """

    @abstractmethod
    def get_sw_sync(self) -> BaseSensorWriteSynchronization:
        """
        Get a synchronization

        This will never be called if dont_do_pathpoints is True
        """



import logging
import typing as tp

import pkg_resources
from OpenSSL import crypto
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from pyasn1.codec.der.decoder import decode
from pyasn1.error import PyAsn1Error
from satella.coding import reraise_as
from satella.coding.structures import Singleton
from satella.files import read_in_file

from smok.basics import Environment
from smok.exceptions import InvalidCredentials

logger = logging.getLogger(__name__)

DEVICE_ID = x509.ObjectIdentifier('1.3.6.1.4.1.55338.0.0')
ENVIRONMENT = x509.ObjectIdentifier('1.3.6.1.4.1.55338.0.1')
# noinspection PyProtectedMember
x509.oid._OID_NAMES[DEVICE_ID] = 'DeviceID'
# noinspection PyProtectedMember
x509.oid._OID_NAMES[ENVIRONMENT] = 'Environment'


def get_root_cert() -> bytes:
    """
    :return: the bytes sequence for SMOK's master CA certificate
    """
    ca_file = pkg_resources.resource_filename('smok', 'certs/root.crt', )
    return read_in_file(ca_file)


def get_rapid_ca_cert() -> bytes:
    """
    :return: the bytes sequence for a PEM-encoded RAPID host signing CA
    """


def get_dev_ca_cert() -> bytes:
    """
    :return: the bytes sequence for SMOK's device signing CA
    """
    ca_file = pkg_resources.resource_filename('smok', 'certs/dev.crt', )
    return read_in_file(ca_file)


@Singleton
class DevRootCertificateStore:
    __slots__ = ('store',)

    def add_certificate(self, name: str):
        ca_file = pkg_resources.resource_filename('smok', 'certs/%s' % (name,))
        cert_pem_data = read_in_file(ca_file)
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_pem_data)
        self.store.add_cert(cert)

    def __init__(self):
        self.store = crypto.X509Store()
        self.add_certificate('root.crt')
        self.add_certificate('dev.crt')
        self.add_certificate('rapid.crt')


DevRootCertificateStore()


def check_if_trusted(cert_data: bytes) -> bool:
    try:
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_data)
    except crypto.Error as e:
        raise InvalidCredentials('problem loading certificate - certificate is invalid') from e
    store_ctx = crypto.X509StoreContext(DevRootCertificateStore().store, cert)
    try:
        store_ctx.verify_certificate()
        return True
    except crypto.Error:
        return False


def get_device_info(cert_data: bytes) -> tp.Tuple[str, Environment]:
    try:
        cert = x509.load_pem_x509_certificate(cert_data, default_backend())
    except ValueError as e:
        raise InvalidCredentials('Error unserializing certificate') from e

    try:
        device_asn1 = cert.extensions.get_extension_for_oid(DEVICE_ID).value.value
    except x509.extensions.ExtensionNotFound as e:
        return InvalidCredentials('DEVICE_ID not found in cert: %s' % (e,))

    try:
        device_id = str(decode(device_asn1)[0])
    except (PyAsn1Error, IndexError) as e:
        return InvalidCredentials('error during decoding DEVICE_ID: %s' % (e,))

    try:
        environment_asn1 = cert.extensions.get_extension_for_oid(ENVIRONMENT).value.value
    except x509.extensions.ExtensionNotFound as e:
        raise InvalidCredentials(str(e))

    try:
        environment = int(decode(environment_asn1)[0])
    except (PyAsn1Error, IndexError, TypeError) as e:
        raise InvalidCredentials('error during decoding environment: %s' % (e,)) from e
    except ValueError as e:
        raise InvalidCredentials('unrecognized environment: %s' % (e,)) from e

    with reraise_as(ValueError, InvalidCredentials):
        return device_id, Environment(environment)



import logging
import time
import typing as tp
import warnings
import weakref
from concurrent.futures import Future

from satella.coding import wraps
from satella.coding.structures import ReprableMixin, OmniHashableMixin
from satella.coding.typing import Number
from satella.instrumentation import Traceback

from .orders import AdviseLevel, Section, ReadOrder, WriteOrder
from .typing import PathpointValueType, ValueOrExcept
from ..basics import StorageLevel
from ..exceptions import OperationFailedError, InstanceNotReady, NotReadedError

logger = logging.getLogger(__name__)


def must_have_device(fun):
    @wraps(fun)
    def inner(self, *args, **kwargs):
        if not self.device:
            raise InstanceNotReady('You need to attach this pathpoint to a device first')
        return fun(self, *args, **kwargs)

    return inner


class Pathpoint(ReprableMixin, OmniHashableMixin):
    """
    Base class for an user-defined pathpoint.

    Note that pathpoint is registered in the device as part of it's creation.

    :param device: device that this pathpoint should be attached to. None is also a valid option.
    :param name: pathpoint name.
    :param storage_level: storage level for this pathpoint
    :param callable_on_change: a callable to be called each time this pathpoint changes value,
        with this pathpoint as it's sole argument. Should this callable return an exception, it will
        be logged as an ERROR along with it's traceback.
    :param read_no_often_than: this pathpoint should be readed at fastest each this seconds.
        Leave it at None (default) to disable this mechanism.

    :ivar name: pathpoint name
    :ivar storage_level: pathpoint's storage level
    :ivar current_timestamp: a timestamp in seconds of the last read
    :ivar current_value: last readed value or an exception instance
    :ivar device: a weak reference to the device
    """
    _HASH_FIELDS_TO_USE = 'name',
    _REPR_FIELDS = 'name', 'storage_level'
    __slots__ = 'name', 'storage_level', 'current_value', 'current_timestamp', 'device', \
                'callable_on_change', 'read_no_often_than', 'last_read'

    def __init__(self, device: tp.Optional['SMOKDevice'], name: str,
                 storage_level: StorageLevel = StorageLevel.TREND,
                 callable_on_change: tp.Optional[tp.Callable[['Pathpoint'], None]] = None,
                 read_no_often_than: tp.Optional[float] = None):
        self.device = weakref.proxy(device)
        self.name = name
        self.read_no_often_than = read_no_often_than
        self.last_read = 0
        self.callable_on_change = callable_on_change
        self.storage_level = storage_level
        # Try to load
        try:
            self.current_timestamp, self.current_value = device.pp_database.get_current_value(name)
        except NotReadedError:
            self.current_timestamp = None
            self.current_value = None
        except OperationFailedError as e:
            self.current_timestamp = e.timestamp
            self.current_value = e

        # noinspection PyProtectedMember
        if device is not None:
            device.register_pathpoint(self)

    def can_read(self) -> bool:
        """
        Called by the executor before a ReadOrder is processed to determine whether this
        pathpoint can be read.

        :return: Can this pathpoint be read?
        """
        if self.read_no_often_than is not None:
            return time.monotonic() - self.last_read > self.read_no_often_than
        return True

    def get_archive(self,
                    starting_at: int,
                    stopping_at: tp.Optional[int] = None) -> tp.Iterator[
        tp.Tuple[int, ValueOrExcept]]:
        """
        Get archive readings.

        This will be readed entirely from the device's data, the server will not be queried

        :param starting_at: timestamp of start, in milliseconds
        :param stopping_at: timestamp of end, in milliseconds, or None for the end of the park
        :return: an iterator of tuple (timestamp in milliseconds,
                                       pathpoint value or exception instance)
        """
        yield from self.device.pp_database.get_archive_data(self.name, starting_at, stopping_at)

    @must_have_device
    def set_new_value(self, *args) -> None:
        """
        May be called asynchronously by user threads to asynchronously update a pathpoint.

        This is also called by the executor thread upon reading a new piece of data.

        You can use it in one of two ways:

        >>> pp.set_new_value(timestamp, value)

        or:

        >>> pp.set_new_value(OperationFailedError(...))

        :param timestamp: new timestamp
        :param value: new value
        """
        if len(args) == 1:
            value, = args
            timestamp = value.timestamp
        else:
            timestamp, value = args
        if self.current_timestamp is not None:
            if self.current_timestamp >= timestamp:
                warnings.warn(
                    'Given lower or equal timestamp (%s) than current one (%s), ignoring' % (
                        timestamp, self.current_timestamp), UserWarning)
                return
        self.device.pp_database.on_new_data(self.name, timestamp, value)
        if self.device.getter is not None:
            self.device.getter.data_to_update.notify()
        self.current_timestamp = timestamp
        self.current_value = value
        if self.callable_on_change is not None:
            try:
                self.callable_on_change(self)
            except Exception as e:
                logger.error('Callback on change for %s failed with %s, ignoring', self.name,
                             Traceback().pretty_format(), exc_info=e)
        self.last_read = time.monotonic()

    def on_read(self, advise: AdviseLevel) -> tp.Optional[tp.Union[PathpointValueType, Future]]:
        """
        Called when there's a request to read this pathpoint.

        This is called from a separate thread spawned by SMOKDevice.

        Note that :meth:`~smok.pathpoint.Pathpoint.can_read` will be checked for before this is
        called.

        This may return a Future, that results in the value readed, or should raise
        OperationFailedError when the read fails, or should directly return the value,
        or should directly raise the exception.

        .. note:: :attr:`current_timestamp` and :attr:`current_value` will be automatically
                  updated, so there's no need for the future to do that.

        :param advise: advise level of this read operation
        :returns: a Future that returns the value of this pathpoint or raises OperationFailedError.
            This Future can also return None, if the value will be set later via
            :meth:`~smokclient.pathpoint.Pathpoint.set_new_value`
        :raises OperationFailedError: operation failed
        """

    def on_write(self, value: PathpointValueType, advise: AdviseLevel) -> tp.Optional[Future]:
        """
        Called when there's a request to write this pathpoint with a particular value

        This is called from a separate thread spawned by SMOKDevice

        The future should raise OperationFailedError when the write fails.
        This can return None if the write won't be executed and should be treated as successful.

        :param value: value to be written
        :param advise: advise level of this read operation
        :returns: a Future that completes successfully if written correctly or excepts if failed
            (you should raise OperationFailedError).
        """

    def on_new_storage_level(self, new_storage_level: StorageLevel) -> None:
        """
        Called when this pathpoint's storage level is updated.

        This is usually invoked by syncing the pathpoints with the server.

        This should complete ASAP, as this is called by a thread that is responsible
        for communication with the server.

        :param new_storage_level: new storage level for this pathpoint
        """
        self.storage_level = new_storage_level

    @must_have_device
    def get(self) -> tp.Tuple[Number, PathpointValueType]:
        """
        Return the current pathpoint value

        :return: a tuple of (last timestamp, when the operation has failed)
        :raises OperationFailedError: when pathpoint's read has failed
        :raises NotReadedError: pathpoint not yet readed
        """
        if self.current_value is None:
            raise NotReadedError()

        if isinstance(self.current_value, OperationFailedError):
            raise self.current_value
        else:
            return self.current_timestamp, self.current_value

    def read(self, advise_level: AdviseLevel = AdviseLevel.ADVISE) -> Section:
        """
        Return an order that reads this pathpoint

        Dispatch them like this:

        >>> my_sd.execute(pp.read(2))

        :param advise_level: required advise level
        :return: a Section, whose execution will refresh the value in the pathpoint
        """
        return Section([ReadOrder(self.name, advise_level)])

    def write(self, value: PathpointValueType, advise_level: AdviseLevel = AdviseLevel.ADVISE,
              stale_after: tp.Optional[float] = None) -> Section:
        """
        Return an order, whose execution will put target value in target pathpoint.

        Dispatch them like this:

        >>> my_sd.execute(pp.write(2))

        :param value: value to write
        :param advise_level: advise level with which to write
        :param stale_after: timestamp, that if given, will prevent the write from being executed
            past it
        :return: a Section, whose execution will put the target value in the pathpoint
        """
        return Section([WriteOrder(self.name, value, advise_level, stale_after)])



import copy
import logging
import typing as tp

from satella.coding import silence_excs, for_argument
from satella.coding.predicates import x
from satella.coding.sequences import choose

from smok.extras.macros_database.base import BaseMacroDatabase
from smok.macro import Macro

logger = logging.getLogger(__name__)


class InMemoryMacroDatabase(BaseMacroDatabase):
    def __init__(self):
        self.macros_to_execute = []  # type: tp.List[Macro]
        self.executions_to_sync = []  # type: tp.List[tp.Tuple[str, int]]

    def set_macros(self, macros: tp.List[Macro]) -> None:
        self.macros_to_execute = macros

    def get_macros(self) -> tp.List[Macro]:
        return self.macros_to_execute

    @silence_excs(ValueError)
    def notify_macro_executed(self, macro_id: str, timestamp: int) -> None:
        macro = choose(x.macro_id == macro_id, self.macros_to_execute)  # throws ValueError
        self.executions_to_sync.append((macro_id, timestamp))
        if not macro:
            self.macros_to_execute.remove(macro)

    @for_argument(returns=copy.copy)
    def get_done_macros(self) -> tp.Iterator[tp.Tuple[str, int]]:
        return self.executions_to_sync

    @silence_excs(ValueError)
    def notify_macro_synced(self, macro_id: str, timestamp: int) -> None:
        self.executions_to_sync.remove((macro_id, timestamp))



from setuptools import find_packages
from distutils.core import setup
from smok import __version__

setup(version=__version__,
      packages=find_packages(include=['smok', 'smok.*', 'ngtt', 'ngtt.*']),
      install_requires=[line.strip() for line in open('requirements.txt', 'r').readlines() if
                        line.strip()],
      )



from satella.coding import Monitor

from .base import BaseSensorWriteDatabase, BaseSensorWriteSynchronization
from ...sensor import SensorWriteEvent


class InMemorySensorWriteDatabase(BaseSensorWriteDatabase, Monitor):
    @Monitor.synchronized
    def on_synced_sw(self, event: SensorWriteEvent):
        self.events.remove(event)

    @Monitor.synchronized
    def on_sync_sw_failed(self, event: SensorWriteEvent):
        pass

    @Monitor.synchronized
    def add_sw(self, event: SensorWriteEvent):
        self.events.add(event)

    @Monitor.synchronized
    def get_sw_sync(self) -> BaseSensorWriteSynchronization:
        return BaseSensorWriteSynchronization(list(self.events), self)

    def __init__(self):
        Monitor.__init__(self)
        self.events = set()



"""
This exports a function to evaluate Python expression that use eval() within SAI5.
"""
import struct
import typing as tp


# fun stuff
def D(real, ref):
    return abs(ref - real)


def d(real, ref):
    return abs(D(real, ref) / real)


def ked(real, ref):
    """
    Return relative error between referential value (real) and measured value (ref)
    real and ref are respective
    returns zero if referential value is below 20
    """
    if real < 20:
        return 0
    return d(real, ref)


def huba505(voltage):
    """
    Given that v is voltage in Volts, return pressure as given by a HUBA 505 transducer

    The transducer is:
    0.5V    -   0 bar
    3.5V    -   4 bar
    """
    pressure = 4 * (voltage - 0.5) / 3
    if pressure > 0:
        return pressure
    else:
        return 0


KTY81 = [
    (-40, 1136),
    (-20, 1250),
    (-10, 1372),
    (0, 1500),
    (10, 1634),
    (20, 1774),
    (25, 1922),
    (30, 2000),
    (40, 2078),
    (50, 2240),
    (60, 2410),
    (70, 2590),
    (80, 2780),
    (90, 2978),
    (100, 3182),
    (110, 3392),
]


def kty81(r):
    """Given that r is resistance in Ohms, convert it to KTY81 centigrades.
    Return bounding value if out of range"""
    if r <= 1136:
        return -40
    if r >= 3276:
        return 94.4

    p_temp, p_resistance = KTY81[0]

    for r_temp, r_resistance in KTY81[1:]:
        if p_resistance < r < r_resistance:
            return (r_temp - p_temp) * (r - p_resistance) / (r_resistance - p_resistance) + p_temp

        p_temp, p_resistance = r_temp, r_resistance


PT100 = [80.31, 82.29, 84.27, 86.25, 88.22, 90.19, 92.16, 94.12, 96.09, 98.04,
         100.0, 101.95, 103.9, 105.85, 107.79, 109.73, 111.67, 113.61, 115.54, 117.47,
         119.4, 121.32, 123.24, 125.16, 127.07, 128.98, 130.89, 132.8, 134.7, 136.6,
         138.5, 140.39, 142.29, 157.31, 175.84, 195.84]


def pt1000(r):
    """Given that r is resistance in Ohms, convert it to PT1000 centigrades"""
    r /= 10.0

    t = -50
    i = 0
    if r > PT100[0]:
        while 250 > t:
            dt = 5 if t < 110 else (50 if t > 110 else 40)
            i += 1
            if r < PT100[i]:
                return t + (r - PT100[i - 1]) * dt / (PT100[i] - PT100[i - 1])
            t += dt
    return t


STRUCT_F = struct.Struct('f')
STRUCT_HH = struct.Struct('>HH')
STRUCT_l = struct.Struct('>l')


def _pack_hh(ho, lo) -> bytes:
    return STRUCT_HH.pack(ho, lo)


def mkflt(ho, lo):
    return STRUCT_F.unpack(_pack_hh(ho, lo))[0]


def mkint32(ho, lo):
    return STRUCT_l.unpack(_pack_hh(ho, lo))[0]


def ecre_eval(expression: str, my_locals=None, args: tp.Union[tp.Any, tp.Tuple[tp.Any]] = ()):
    """
    Our limited form of secure Python evaluation.

    Note that this is not really secure against malicious code. See
    https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html for details.

    Always trust the code you execute!

    :param expression: expression to evaluate
    :param my_locals: locals to use. Additionally, locals from the ecre_eval runtime environment
        will be added
    :param args: args to use. They will be available in the expression as v0, v1, v2..
        If a single non-sequence value is passed, it will be available as v0.
    :return: evaluated value
    """

    my_locals = my_locals or {}
    if not isinstance(args, tp.Iterable):
        my_locals['v0'] = args
    else:
        for index, value in enumerate(args):
            my_locals['v%s' % (index, )] = value

    my_locals.update({'d': d, 'D': D, 'ked': ked, 'pt1000': pt1000, 'huba505': huba505,
                      'negz': lambda x: 0 if x < 0 else x,
                      'mkflt': mkflt,
                      'str': str,
                      'kty81': kty81,
                      'mkint32': mkint32})
    return eval(expression, {}, my_locals)



from .base import BaseMetadataDatabase
from .in_memory import InMemoryMetadataDatabase
from .pickling import PicklingMetadataDatabase

__all__ = ['BaseMetadataDatabase', 'InMemoryMetadataDatabase', 'PicklingMetadataDatabase']
